@@ -6,8 +6,8 @@ from scipy.optimize import OptimizeResult
 from scipy.optimize import minimize as sp_minimize
 from sklearn.base import is_regressor
 from sklearn.ensemble import GradientBoostingRegressor
-from sklearn.externals.joblib import dump as dump_
-from sklearn.externals.joblib import load as load_
+from joblib import dump as dump_
+from joblib import load as load_
 
 from .learning import ExtraTreesRegressor
 from .learning import GaussianProcessRegressor
@@ -255,8 +255,8 @@ def has_gradients(estimator):
     estimator: sklearn BaseEstimator instance.
     """
     tree_estimators = (
-            ExtraTreesRegressor, RandomForestRegressor,
-            GradientBoostingQuantileRegressor
+        ExtraTreesRegressor, RandomForestRegressor,
+        GradientBoostingQuantileRegressor
     )
 
     # cook_estimator() returns None for "dummy minimize" aka random values only
@@ -277,7 +277,7 @@ def has_gradients(estimator):
     return not categorical_gp
 
 
-def cook_estimator(base_estimator, space=None, **kwargs):
+def cook_estimator(base_estimator, space=None, length_scale_bounds=None, length_scale=None, **kwargs):
     """
     Cook a default estimator.
 
@@ -299,10 +299,14 @@ def cook_estimator(base_estimator, space=None, **kwargs):
     * `space` [Space instance]:
         Has to be provided if the base_estimator is a gaussian process.
         Ignored otherwise.
-
+    * `length_scale_bounds` [list of tuples]:
+        the length scale bounds for the matern kernel
+    * `length_scale_bounds` [list of floats]:
+        the length scales for the Matern or Hamming kernel
     * `kwargs` [dict]:
         Extra parameters provided to the base_estimator at init time.
     """
+
     if isinstance(base_estimator, str):
         base_estimator = base_estimator.upper()
         if base_estimator not in ["GP", "ET", "RF", "GBRT", "DUMMY"]:
@@ -322,19 +326,33 @@ def cook_estimator(base_estimator, space=None, **kwargs):
         else:
             raise ValueError("Expected a Space instance, not None.")
 
-        cov_amplitude = ConstantKernel(1.0, (0.01, 1000.0))
+        cov_amplitude = ConstantKernel(1.0, (0.01, 1000))
+
+        if not length_scale:
+            length_scale = np.ones(n_dims)
+        if not length_scale_bounds:
+            length_scale_bounds = [(0.1, 1)] * n_dims
+
+        # Transform lengthscale bounds:
+        length_scale_bounds_transformed = []
+        length_scale_transformed = []
+        for i in range(len(space.dimensions)):
+            for j in range(space.dimensions[i].transformed_size):
+                length_scale_bounds_transformed.append(length_scale_bounds[i])
+                length_scale_transformed.append(length_scale[i])
+
         # only special if *all* dimensions are categorical
         if is_cat:
-            other_kernel = HammingKernel(length_scale=np.ones(n_dims))
+            other_kernel = HammingKernel(length_scale=length_scale_transformed)
         else:
             other_kernel = Matern(
-                length_scale=np.ones(n_dims),
-                length_scale_bounds=[(0.01, 100)] * n_dims, nu=2.5)
+                length_scale=length_scale_transformed,
+                length_scale_bounds=length_scale_bounds_transformed, nu=2.5)
 
         base_estimator = GaussianProcessRegressor(
             kernel=cov_amplitude * other_kernel,
             normalize_y=True, noise="gaussian",
-            n_restarts_optimizer=2)
+            n_restarts_optimizer=4)
     elif base_estimator == "RF":
         base_estimator = RandomForestRegressor(n_estimators=100,
                                                min_samples_leaf=3)
@@ -491,13 +509,13 @@ def normalize_dimensions(dimensions):
                     Real(dimension.low, dimension.high, dimension.prior,
                          name=dimension.name,
                          transform="normalize")
-                    )
+                )
             elif isinstance(dimension, Integer):
                 transformed_dimensions.append(
                     Integer(dimension.low, dimension.high,
                             name=dimension.name,
                             transform="normalize")
-                    )
+                )
             else:
                 raise RuntimeError("Unknown dimension type "
                                    "(%s)" % type(dimension))
